import asyncio
import functools
import logging
from datetime import datetime
from pathlib import Path
from typing import AsyncIterator, Awaitable, Callable, OrderedDict, ParamSpec, TypeVar

from pyrogram import Client as TelegramClient
from pyrogram.client import Cache
from pyrogram.errors import FloodWait, PeerIdInvalid
from pyrogram.file_id import FileId
from pyrogram.raw.functions.messages import EditChatAdmin, GetAvailableReactions
from pyrogram.raw.types import (
    InputPeerChat,
    InputPeerUser,
    PeerUser,
    ReactionCustomEmoji,
    ReactionEmoji,
    UserEmpty,
)
from pyrogram.raw.types import User as RawUser
from pyrogram.raw.types.messages import AvailableReactions
from pyrogram.types import Chat, Message, Update, User
from slidge import global_config
from slixmpp.exceptions import XMPPError

from .reactions import ReactionsStore

P = ParamSpec("P")
R = TypeVar("R")
WrappedMethod = Callable[P, R]

AVATAR_DOWNLOAD_SLEEP = 15  # seconds between avatar downloads
MAX_FLOOD_ATTEMPTS = 10


class InvalidUser:
    pass


class InvalidUserException(Exception):
    pass


class Client(TelegramClient):
    message_cache: "MessageCache"

    def __init__(self, name: str) -> None:
        super().__init__(name, workdir=str(global_config.HOME_DIR))

        self._available_reactions: set[str] | None = None
        self.log = logging.getLogger(f"Telegram:{name}")

        self.get_chat = handle_flood(self.get_chat)  # type: ignore
        self.get_contacts = handle_flood(self.get_contacts)  # type: ignore
        self.get_users = handle_flood(self.get_users)  # type: ignore
        self.download_media = handle_flood(self.download_media)  # type: ignore
        self.get_chat_member = handle_flood(self.get_chat_member)  # type: ignore
        self._download_avatar_lock = asyncio.Lock()
        self._get_user_lock = asyncio.Lock()
        self._users_to_get = list[int]()

        self.on_raw_update(group=0)(self._on_raw)  # type:ignore
        self.on_edited_message(group=1)(self._on_edited_message)  # type: ignore

        self._reactions = ReactionsStore(name)
        # We cache raw users because they seem to have a measurably
        # lower memory footprint than pyrogram's "rich" Users.
        self._user_cache: OrderedDict[int, RawUser | InvalidUser] = LimitedSizeDict(
            1_000
        )
        self.message_cache = MessageCache(10_000)

        self._reaction_handler: (
            Callable[[Message, int, str | None], Awaitable[None]] | None
        ) = None

    def is_me(self, user: User | int) -> bool:
        # we need this because is_self is not always set
        if isinstance(user, User):
            if user.is_self is not None:
                return user.is_self
            user_id = user.id
        else:
            user_id = user
        assert self.me is not None
        return self.me.id == user_id

    @property
    def download_path(self):
        # the trailing slash is needed by pyrogram
        return str(global_config.HOME_DIR.absolute() / "telegram_downloads") + "/"

    async def download_avatar(self, file_id: str) -> Path | None:
        async with self._download_avatar_lock:
            await asyncio.sleep(AVATAR_DOWNLOAD_SLEEP)
            path = await self.download_media(file_id, self.download_path)
            if path is None:
                return None
            assert isinstance(path, str)
            return Path(path)

    def get_downloader(self, file_id_str: str) -> AsyncIterator[bytes] | None:
        file_id_obj = FileId.decode(file_id_str)
        return self.get_file(file_id_obj)

    async def get_available_reactions(self) -> AvailableReactions:
        rpc = GetAvailableReactions(hash=0)
        return await self.invoke(rpc)

    async def available_reactions(self) -> set[str]:
        if self._available_reactions is None:
            available_reactions = await self.get_available_reactions()
            self._available_reactions = {
                r.reaction for r in available_reactions.reactions
            }
        return self._available_reactions

    def on_reaction(self, callback):
        self._reaction_handler = callback

    async def _on_edited_message(self, _tg: TelegramClient, message: Message):
        if (
            message.raw.reactions is not None
            and message.raw.reactions.recent_reactions is not None
        ):
            await self._on_reaction(message)

    async def _on_reaction(self, message: Message):
        if self._reaction_handler is None:
            return

        assert message.raw.reactions.recent_reactions is not None
        assert self.me is not None

        new_reacters = set[tuple[int, str]]()
        for reaction in message.raw.reactions.recent_reactions:
            if reaction.my:
                user_id = self.me.id
            elif isinstance(reaction.peer_id, PeerUser):
                user_id = reaction.peer_id.user_id
            else:
                continue

            if isinstance(reaction.reaction, ReactionEmoji):
                new_reacters.add((user_id, reaction.reaction.emoticon))
            elif isinstance(reaction.reaction, ReactionCustomEmoji):
                new_reacters.add((user_id, "❓"))
            else:
                self.log.warning("Unknown reaction: %s", reaction.reaction)

        old_reacters = set(self._reactions.get(message))
        self._reactions.set(message, new_reacters)
        self.log.debug("Old reacters: %s", old_reacters)
        self.log.debug("New reacters: %s", new_reacters)

        old_all_reacters = {x[0] for x in old_reacters}
        new_all_reacters = {x[0] for x in new_reacters}
        for unreacter_id in old_all_reacters - new_all_reacters:
            await self._reaction_handler(message, unreacter_id, None)
        for reacter_id, emoji in new_reacters - old_reacters:
            await self._reaction_handler(message, reacter_id, emoji)

    async def _on_raw(
        self, _tg, update: Update, _users: dict[int, User], _chats: dict[int, Chat]
    ):
        async with self._get_user_lock:
            self.__update_user_cache(update)

    async def get_user(self, user_id: int) -> User:
        """
        Get a telegram user from local cache or through a RPC if needed.

        :param user_id:
        :return:
        """
        async with self._get_user_lock:
            cached_raw = self._user_cache.get(user_id)
            if cached_raw is invalid_user:
                raise InvalidUserException(f"{user_id} was cached and is invalid")
            if cached_raw is not None:
                self.log.debug("user was cached! YAY!")
                # noinspection PyProtectedMember
                cached_user = User._parse(self, cached_raw)  # type:ignore
                assert cached_user is not None
                return cached_user
        self.log.debug("user %s was not cached! damn!", user_id)
        try:
            user = await self.get_users(user_id)
        except PeerIdInvalid:
            self._user_cache[user_id] = invalid_user
            self.log.debug("user %s could not be resolved", user_id)
            raise InvalidUserException(f"{user_id} raised PeerIdInvalid")
        assert isinstance(user, User)
        return user

    async def invoke(self, *a, **k):
        r = await super().invoke(*a, **k)
        self.__update_user_cache(r)
        return r

    def __update_user_cache(self, raw_obj):
        raw_users: list[RawUser] = getattr(raw_obj, "users", [])
        for raw_user in raw_users:
            if isinstance(raw_user, UserEmpty):
                continue
            self._user_cache[raw_user.id] = raw_user

    async def edit_chat_admin(
        self, chat_id: int | str, user_id: int | str, is_admin: bool
    ) -> bool:
        chat = await self.resolve_peer(chat_id)
        if not isinstance(chat, InputPeerChat):
            self.log.warning("Tried to set admin of a wrong peer type: %s", type(chat))
            return False
        user = await self.resolve_peer(user_id)
        if not isinstance(user, InputPeerUser):
            self.log.warning("Tried to make admin but wrong peer type: %s", type(user))
            return False
        return await self.invoke(
            EditChatAdmin(
                chat_id=chat.chat_id,
                user_id=user,  # type:ignore
                is_admin=is_admin,
            )
        )


class LimitedSizeDict(OrderedDict):
    def __init__(self, size: int, *args, **kwargs):
        self._size = size
        super().__init__(*args, **kwargs)
        self._check_size_limit()

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._check_size_limit()

    def update(self, *args, **kwargs):
        super().update(*args, **kwargs)
        self._check_size_limit()

    def _check_size_limit(self):
        while len(self) > self._size:
            self.popitem(last=False)


class MessageCache(Cache):
    # we need to subclass the default message cache because _on_tg_deleted_msg
    # comes with message IDs only, and we need to know which chat they actually
    # belong too

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._chat_by_message_ids = LimitedSizeDict(10_000)

    def __setitem__(self, key: tuple[int, int], value: Message) -> None:
        # tuple = [chat_id, message_id]
        super().__setitem__(key, value)
        self._chat_by_message_ids[value.id] = value

    def get_by_message_id(self, message_id: int) -> Message:
        return self._chat_by_message_ids.get(message_id)  # type:ignore

    def remove_chat(self, chat_id: int) -> None:
        # a bit hacky, but works. maybe perf will be an issue eventually
        new = {
            message_id: message
            for message_id, message in self._chat_by_message_ids.items()
            if message.chat.id != chat_id
        }
        self._chat_by_message_ids = LimitedSizeDict(10_000, new)


invalid_user = InvalidUser()


def handle_flood(func: WrappedMethod) -> WrappedMethod:
    @functools.wraps(func)
    async def wrapped(*a, **kw):
        start = datetime.now()
        for i in range(MAX_FLOOD_ATTEMPTS):
            try:
                return await func(*a, **kw)
            except FloodWait as e:
                log.warning(
                    "Flood in %s(%s %s) (%s), sleep for %s seconds",
                    func.__name__,
                    a,
                    kw,
                    start,
                    e.value,
                )
                await asyncio.sleep(e.value + i)
        raise XMPPError("internal-server-error", "Telegram flood")

    return wrapped


log = logging.getLogger(__name__)



import asyncio
import logging
import os
from typing import TYPE_CHECKING

from pyrogram.enums import ChatType
from pyrogram.types import (
    Animation,
    Audio,
    Document,
    Message,
    Photo,
    Sticker,
    Thumbnail,
    Video,
    VideoNote,
    Voice,
    WebPageEmpty,
)
from slidge.core.mixins.message import ContentMessageMixin
from slidge.util.types import LinkPreview, MessageReference
from slixmpp.exceptions import XMPPError

from . import config
from .telegram import handle_flood
from .text_entities import entities_to_xep_0393

if TYPE_CHECKING:
    from .gateway import Gateway
    from .session import Session

TgMediaTypes = (
    Audio
    | Document
    | Photo
    | Sticker
    | Animation
    | Voice
    | Video
    | VideoNote
    | Thumbnail
)

MSG_POLL = "/me sent a poll but this is not supported by slidgram yet"


class TelegramMessageSenderMixin(ContentMessageMixin):
    xmpp: "Gateway"
    session: "Session"
    log: logging.Logger

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.send_file = handle_flood(self.send_file)  # type:ignore
        self._convert_args = [
            "--height",
            str(config.CONVERT_STICKERS_SIZE),
            "--width",
            str(config.CONVERT_STICKERS_SIZE),
            "--fps",
            str(config.CONVERT_STICKERS_FPS),
        ]

    @property
    def tg(self):
        return self.session.tg

    @property
    def contacts(self):
        return self.session.contacts

    @property
    def bookmarks(self):
        return self.session.bookmarks

    async def send_tg_msg(
        self, message: Message, carbon=False, correction=False, archive_only=False
    ) -> None:
        if message.poll is not None:
            await self.__send_text(message, carbon, correction, archive_only, MSG_POLL)
        elif message.media is not None and message.web_page_preview is None:
            await self._send_media(message, carbon, correction, archive_only)
        else:
            await self.__send_text(message, carbon, correction, archive_only)

    async def __send_text(
        self,
        message: Message,
        carbon=False,
        correction=False,
        archive_only=False,
        text: str | None = None,
    ):
        self.send_text(
            self._to_message_styling(message) if text is None else text,
            message.id,
            reply_to=await self._get_reply_to(message.reply_to_message),
            carbon=carbon,
            correction=correction,
            when=message.date,
            archive_only=archive_only,
            link_previews=_get_link_previews(message),
        )

    async def _send_media(
        self, message: Message, carbon: bool, correction=False, archive_only=False
    ) -> None:
        if (
            message.sticker is not None
            and message.sticker.is_animated
            and config.CONVERT_STICKERS
        ):
            try:
                await self.__send_sticker(message, carbon, correction, archive_only)
            except Exception as e:
                self.log.error("Could not convert stickers.", exc_info=e)
            else:
                return

        media = _get_media(message)
        if media is None:
            self.log.warning("Could not determine media in %s", message)
            await self.__send_text(
                message,
                carbon,
                correction,
                archive_only,
                f"Unsupported media type: {message.media}",
            )
            return
        if message.caption is None:
            caption = None
        else:
            caption = self._to_message_styling_caption(message)
        if media.file_size > config.ATTACHMENT_MAX_SIZE:
            text = f"{media} (larger than {config.ATTACHMENT_MAX_SIZE})"
            if message.text:
                text += f"\n{self._to_message_styling(message)}"
            if caption:
                text += f"\n{caption}"
            await self.__send_text(
                message,
                carbon,
                correction,
                archive_only,
                text,
            )
            return
        self.log.debug("Downloading %s", media.file_id)
        downloader = self.tg.get_downloader(media.file_id)
        if downloader is None:
            self.log.warning("Could not download %s", media)
            return
        await self.send_file(
            None,
            message.id,
            async_data_stream=downloader,
            file_name=getattr(media, "file_name", None),
            legacy_file_id=media.file_unique_id,
            caption=caption,
            reply_to=await self._get_reply_to(message.reply_to_message),
            carbon=carbon,
            correction=correction,
            when=message.date,
            archive_only=archive_only,
            link_previews=_get_link_previews(message),
            content_type=(
                "image/jpeg"
                if isinstance(
                    media, Photo
                )  # no mime_type attribute for Photos, but always JPEG
                else getattr(media, "mime_type", None)
            ),
        )

    async def __send_sticker(
        self, message: Message, carbon: bool, correction=False, archive_only=False
    ):
        sticker = message.sticker
        sticker_id = sticker.file_unique_id
        webm_path = (self.xmpp.stickers_dir / sticker_id).with_suffix(".tgs.webm")

        if not webm_path.exists():
            tgs_filename = (self.xmpp.stickers_dir / sticker_id).with_suffix(".tgs")
            downloader = self.tg.get_downloader(sticker.file_id)
            with tgs_filename.open("wb") as fp:
                async for chunk in downloader:
                    fp.write(chunk)
            self.log.debug("Converting sticker %s to video", sticker.file_id)
            async with _conversion_lock:
                proc = await asyncio.create_subprocess_exec(
                    config.CONVERT_STICKERS_EXECUTABLE,
                    str(tgs_filename),
                    *self._convert_args,
                )
                await proc.communicate()
            self.log.debug("Conversion finished with return code: %s", proc.returncode)

        await self.send_file(
            legacy_msg_id=message.id,
            file_path=webm_path,
            legacy_file_id="sticker-" + sticker_id,
            reply_to=await self._get_reply_to(message.reply_to_message),
            carbon=carbon,
            correction=correction,
            when=message.date,
            archive_only=archive_only,
            content_type="video/webm",
        )

    async def _get_reply_to(self, message: Message | None) -> MessageReference | None:
        if message is None:
            return None

        if message.from_user is not None:
            if self.tg.is_me(message.from_user):
                author = "user"
            else:
                if message.chat.type in (ChatType.PRIVATE, ChatType.BOT):
                    try:
                        author = await self.contacts.by_legacy_id(message.from_user.id)
                    except XMPPError as e:
                        # deleted/banned user?
                        if e.condition == "item-not-found":
                            author = None
                        else:
                            raise
                else:
                    muc = await self.bookmarks.by_legacy_id(message.chat.id)
                    try:
                        author = await muc.get_participant_by_legacy_id(
                            message.from_user.id
                        )
                    except XMPPError as e:
                        # deleted/banned user?
                        if e.condition == "item-not-found":
                            author = None
                        else:
                            raise
        elif message.sender_chat is not None or message.chat.type == ChatType.CHANNEL:
            muc = await self.bookmarks.by_legacy_id(message.chat.id)
            author = muc.get_system_participant()
        else:
            self.log.warning("Referenced message author not understood: %s", message)
            author = None

        return MessageReference(
            message.id,
            author,  # type:ignore
            self._to_message_styling(message),
        )

    def _to_message_styling(self, message: Message) -> str:
        assert self.tg.me is not None
        return entities_to_xep_0393(
            message.text, message.entities, self.tg.me.id, self.bookmarks.user_nick
        )

    def _to_message_styling_caption(self, message: Message) -> str:
        assert self.tg.me is not None
        return entities_to_xep_0393(
            message.caption,
            message.caption_entities,
            self.tg.me.id,
            self.bookmarks.user_nick,
        )


def _get_link_previews(message: Message) -> list[LinkPreview] | None:
    if message.web_page_preview is None:
        return None

    page = message.web_page_preview.webpage

    if isinstance(page, WebPageEmpty):
        return None

    return [
        LinkPreview(
            about=page.description,
            title=page.title,
            description=page.description,
            url=page.url,
            image=None,
            type=page.type,
            site_name=page.site_name,
        )
    ]


def _get_media(message: Message) -> TgMediaTypes | None:
    if message.sticker is not None:
        if message.sticker.is_animated and message.sticker.thumbs:
            return message.sticker.thumbs[0]
        if message.sticker.is_video:
            return message.sticker
    for name in _MEDIAS:
        media = getattr(message, name, None)
        if media is not None:
            return media
    return None


_MEDIAS = (
    "audio",
    "document",
    "photo",
    "sticker",
    "animation",
    "video",
    "voice",
    "video_note",
    "new_chat_photo",
)

cpu_count = os.cpu_count()
if cpu_count is None or cpu_count <= 2:
    _conversion_lock: asyncio.Lock | asyncio.Semaphore = asyncio.Lock()
else:
    _conversion_lock = asyncio.Semaphore(cpu_count - 1)



from typing import TYPE_CHECKING

from slidge import FormField
from slidge.command import Command, CommandAccess, Form
from slidge.command.categories import GROUPS

if TYPE_CHECKING:
    from .session import Session


class JoinPublicChat(Command):
    NAME = "🚪 Join a telegram chat"
    HELP = "Join a public channel, private group or supergroup"
    NODE = CHAT_COMMAND = "join-chat"
    ACCESS = CommandAccess.USER_LOGGED
    INSTRUCTIONS = "Use a tg:// URI or a or a https://t.me URL to join a group"
    CATEGORY = GROUPS

    async def run(self, _session, _ifrom, *_args):
        return Form(
            title=self.NAME,
            instructions=self.INSTRUCTIONS,
            fields=[FormField("query", label="Username, tg:// or t.me URL")],
            handler=self.finish,  # type:ignore
        )

    @staticmethod
    async def finish(form_values: dict, session: "Session", _ifrom):
        chat_name: str = form_values["query"]
        if chat_name.startswith("http://"):
            chat_name = "https://" + chat_name[7:]

        # The /s/ part is for web preview of telegram chats and is not accepted by pyrofork's API
        chat_name = chat_name.replace("https://t.me/s/", "https://t.me/")

        chat = await session.tg.join_chat(chat_name)
        muc = await session.bookmarks.by_legacy_id(chat.id)
        return f"You can now join '{chat.title}' at xmpp:{muc.jid}?join"



from typing import TYPE_CHECKING, Sequence

import sqlalchemy as sa
from pyrogram.types import Message
from sqlalchemy import orm

if TYPE_CHECKING:
    from .session import Session


class Base(orm.DeclarativeBase):
    pass


class Reaction(Base):
    __tablename__ = "reaction"
    __table_args = sa.Index("client_name", "chat_id", "message_id")

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    client_name: orm.Mapped[str] = orm.mapped_column()

    chat_id: orm.Mapped[int] = orm.mapped_column()
    message_id: orm.Mapped[int] = orm.mapped_column()

    user_id: orm.Mapped[int] = orm.mapped_column()
    emoji: orm.Mapped[str] = orm.mapped_column()


class ReactionsMixin:
    session: "Session"

    legacy_id: int
    REACTIONS_SINGLE_EMOJI = True

    async def available_emojis(self, _msg_id=None) -> set[str] | None:
        return await self.session.tg.available_reactions()


class ReactionsStore:
    __slots__ = "_name"

    def __init__(self, name: str):
        self._name = name

    def get(self, message: Message) -> Sequence[tuple[int, str]]:
        with orm.Session(engine) as session:  # noqa: F821
            return session.execute(  # type:ignore
                sa.select(Reaction.user_id, Reaction.emoji)
                .where(Reaction.client_name == self._name)
                .where(Reaction.chat_id == message.chat.id)
                .where(Reaction.message_id == message.id)
            ).all()

    def set(self, message: Message, reactions: set[tuple[int, str]]) -> None:
        with orm.Session(engine) as session:  # noqa: F821
            session.execute(
                sa.delete(Reaction)
                .where(Reaction.client_name == self._name)
                .where(Reaction.chat_id == message.chat.id)
                .where(Reaction.message_id == message.id)
            )

            for user_id, emoji in reactions:
                session.add(
                    Reaction(
                        client_name=self._name,
                        chat_id=message.chat.id,
                        message_id=message.id,
                        user_id=user_id,
                        emoji=emoji,
                    )
                )

            session.commit()


engine: sa.Engine



from typing import Optional

_api_txt = "\nIf you dont set it, users will have to enter their own on registration."

API_ID: Optional[int] = None
API_ID__DOC = "Telegram app api_id, obtained at https://my.telegram.org/apps" + _api_txt

API_HASH: Optional[str] = None
API_HASH__DOC = (
    "Telegram app api_hash, obtained at https://my.telegram.org/apps" + _api_txt
)

REGISTRATION_AUTH_CODE_TIMEOUT: int = 60
REGISTRATION_AUTH_CODE_TIMEOUT__DOC = (
    "On registration, users will be prompted for a 2FA code they receive "
    "on other telegram clients."
)

GROUP_HISTORY_MAXIMUM_MESSAGES = 50
GROUP_HISTORY_MAXIMUM_MESSAGES__DOC = (
    "The number of messages to fetch from a group history. "
    "These messages and their attachments will be fetched on slidge startup."
)

ATTACHMENT_MAX_SIZE: int = 10 * 1024**2
ATTACHMENT_MAX_SIZE__DOC = (
    "Maximum file size (in bytes) to download from telegram automatically/"
)

BIG_AVATARS = False
BIG_AVATARS__DOC = (
    "Fetch contact avatars in high-resolution (640x640) instead of the "
    "default 160x160. NB: slidge core main config AVATAR_SIZE still applies."
)

CONVERT_STICKERS = True
CONVERT_STICKERS__DOC = (
    "Convert incoming animated stickers to webm videos. "
    "Requires lottie_to_webm.sh in $PATH, cf <https://github.com/ed-asriyan/lottie-converter>, "
    "along with FFMPEG."
)

CONVERT_STICKERS_EXECUTABLE = "lottie_to_webm.sh"
CONVERT_STICKERS_EXECUTABLE__DOC = "Path to the TGS/webm converter executable."

CONVERT_STICKERS_SIZE = 128
CONVERT_STICKERS_SIZE__DOC = "Width and height video stickers."

CONVERT_STICKERS_FPS = 60
CONVERT_STICKERS_FPS__DOC = "Framerate of the video stickers"



"""
Converts telegram formatted text to XEP-0393 (message styling) strings.
"""

import logging

from pyrogram.enums import MessageEntityType
from pyrogram.types import MessageEntity
from slidge.util.types import Mention
from slidge_style_parser import format_for_telegram  # type:ignore

_STYLING_SURROUNDS = {
    MessageEntityType.ITALIC: "_".encode("utf-16-le"),
    MessageEntityType.BOLD: "*".encode("utf-16-le"),
    MessageEntityType.STRIKETHROUGH: "~".encode("utf-16-le"),
    MessageEntityType.CODE: "`".encode("utf-16-le"),
}
CODE_BLOCK_TERM = "\n```\n".encode("utf-16-le")
NEW_LINE_UTF_16 = "\n".encode("utf-16-le")


def entities_to_xep_0393(
    text: str,
    entities: list[MessageEntity],
    user_id: int | None = None,
    user_nick: str | None = None,
):
    if not entities:
        return text

    # when there is nesting, telegram split entities, but we want to
    # avoid "_this__*is bold*__nested in italic_"
    #
    # # the split similar entities are not guaranteed to be consecutive,
    # # so we first regroup by ID

    try:
        # then we merge and sort by offset because our converter requires that
        entities = sorted(merge_consecutive_entities(entities), key=lambda x: x.offset)

        text_utf16 = text.encode("utf-16-le")
        for e in entities:
            e.offset *= 2
            e.length *= 2
        res_utf16 = entities_to_xep_0393_utf_16(
            text_utf16, entities, user_id, user_nick
        )

        return res_utf16.decode("utf-16-le")
    except Exception as e:
        # let's log it all so we understand why this sometimes happen
        log.exception(
            "Conversion of '%s' with entities '%s' to message styling failed, "
            "falling back to basic text content.",
            text,
            entities,
            exc_info=e,
        )
        return text


def entities_to_xep_0393_utf_16(
    text: bytes,
    entities: list[MessageEntity],
    user_id: int | None = None,
    user_nick: str | None = None,
):
    result = b""
    index = 0
    while entities:
        entity = entities.pop(0)

        offset = entity.offset
        length = entity.length
        end = offset + length

        if (
            entity.type == MessageEntityType.CODE
            and NEW_LINE_UTF_16 in text[offset:end]
        ):
            # telegram allows new lines in preformatted blocks, but
            # XEP-0393 requires ``` instead of ` for that
            entity.type = MessageEntityType.PRE

        before = text[index:offset]
        result += to_xep_0393(before)

        inside_entities = []
        while entities and entities[0].offset < end:
            inside_entities.append(entities.pop(0))

        for inside_entity in inside_entities:
            inside_entity.offset -= offset

        match = text[offset:end]
        match_md = entities_to_xep_0393_utf_16(match, inside_entities)
        result += to_xep_0393(match_md, entity, user_id, user_nick)
        index = end

    after = text[index:]
    result += to_xep_0393(after)

    return result


def to_xep_0393(
    t: bytes,
    entity: MessageEntity | None = None,
    user_id: int | None = None,
    user_nick: str | None = None,
):
    if not entity:
        return t

    type_ = entity.type
    surround = _STYLING_SURROUNDS.get(type_)
    if surround:
        return surround + t + surround

    if type_ == MessageEntityType.PRE:
        return f"\n```{entity.language}\n".encode("utf-16-le") + t + CODE_BLOCK_TERM

    if type_ in (MessageEntityType.TEXT_LINK, MessageEntityType.URL):
        if entity.url:
            return t + f"<{entity.url}>".encode("utf-16-le")
        return t

    if (
        type_ in (MessageEntityType.TEXT_MENTION, MessageEntityType.MENTION)
        and entity.user is not None
        and entity.user.id == user_id
        and user_nick
    ):
        return user_nick.encode("utf-16-le")

    return t


def merge_consecutive_entities(entities: list[MessageEntity]):
    result = []
    i = 0
    while i < len(entities):
        j = i
        add = 0
        while j < len(entities) - 1:
            e1 = entities[j]
            e2 = entities[j + 1]
            if e1.type == e2.type and e1.offset + e1.length == e2.offset:
                j += 1
                add += e2.length
            else:
                break
        entities[i].length += add
        result.append(entities[i])
        i = j + 1

    return result


async def styling_to_entities(
    text: str, mentions: list[Mention] | None = None
) -> tuple[str, list[MessageEntity]]:
    if mentions is None:
        mentions = []
    text, blocks = format_for_telegram(
        text, [(m.contact.name, m.start, m.end) for m in mentions]
    )
    entities = []
    for formatting, offset, length, lang in blocks:
        if formatting == "mention":
            entities.append(
                MessageEntity(
                    type=MessageEntityType.TEXT_MENTION,
                    offset=offset,
                    length=length,
                    user=await mentions.pop(0).contact.get_tg_user(),  # type:ignore
                )
            )
        elif formatting in ("code", "pre"):
            entities.append(
                MessageEntity(
                    type=MessageEntityType.PRE if lang else MessageEntityType.CODE,
                    offset=offset,
                    length=length,
                    language=lang or None,  # type:ignore
                )
            )
        else:
            entities.append(
                MessageEntity(
                    type=PARSER_TO_ENTITY[formatting],
                    offset=offset,
                    length=length,
                )
            )
    return text, entities


PARSER_TO_ENTITY = {
    "italics": MessageEntityType.ITALIC,
    "bold": MessageEntityType.BOLD,
    "strikethrough": MessageEntityType.STRIKETHROUGH,
    "spoiler": MessageEntityType.SPOILER,
}

log = logging.getLogger(__name__)



from pyrogram.types import MessageEntity
from pyrogram.enums import MessageEntityType

from slidgram.text_entities import merge_consecutive_entities, entities_to_xep_0393


def test_parse_entities():
    assert (
        entities_to_xep_0393(
            "Bon erfd dsa sdf",
            [
                MessageEntity(offset=4, length=4, type=MessageEntityType.ITALIC),
                MessageEntity(offset=9, length=3, type=MessageEntityType.BOLD),
            ],
        )
        == "Bon _erfd_ *dsa* sdf"
    )


def test_parse_nested_entities():
    assert (
        entities_to_xep_0393(
            "Bon erfd dsa sdf",
            [
                MessageEntity(offset=3, length=8, type=MessageEntityType.BOLD),
                MessageEntity(offset=4, length=4, type=MessageEntityType.ITALIC),
            ],
        )
        == "Bon* _erfd_ ds*a sdf"
    )


def test_code_block():
    assert (
        entities_to_xep_0393(
            "Example:\ndef prout():\n    print('P*O*T')\nBABY!!",
            [
                MessageEntity(
                    offset=9, length=31, type=MessageEntityType.CODE, language="python"
                ),
                MessageEntity(
                    offset=41, length=4, type=MessageEntityType.STRIKETHROUGH
                ),
            ],
        )
        == "Example:\n\n```python\ndef prout():\n    print('P*O*T')\n```\n\n~BABY~!!"
    )


def test_link():
    assert (
        entities_to_xep_0393(
            "Click this link.",
            [
                MessageEntity(
                    offset=11,
                    length=4,
                    type=MessageEntityType.URL,
                    url="http",
                ),
            ],
        )
        == "Click this link<http>."
    )


def test_merge():
    assert merge_consecutive_entities(
        [
            MessageEntity(offset=2, length=4, type=MessageEntityType.BOLD),
            MessageEntity(offset=6, length=1, type=MessageEntityType.BOLD),
            MessageEntity(offset=7, length=1, type=MessageEntityType.BOLD),
            MessageEntity(offset=12, length=15, type=MessageEntityType.BOLD),
            MessageEntity(offset=27, length=20, type=MessageEntityType.ITALIC),
        ]
    ) == [
        MessageEntity(offset=2, length=6, type=MessageEntityType.BOLD),
        MessageEntity(offset=12, length=15, type=MessageEntityType.BOLD),
        MessageEntity(offset=27, length=20, type=MessageEntityType.ITALIC),
    ]



import functools
from typing import TYPE_CHECKING, Any, Callable, ParamSpec, TypeVar

from pyrogram.errors import (
    AuthKeyUnregistered,
    BadRequest,
    Forbidden,
    ReactionInvalid,
    RPCError,
    Unauthorized,
)
from slixmpp.exceptions import XMPPError
from slixmpp.types import ErrorConditions

from .telegram import InvalidUserException

if TYPE_CHECKING:
    from .session import Session

P = ParamSpec("P")
R = TypeVar("R")
WrappedMethod = Callable[P, R]


_ERROR_MAP: dict[Any, ErrorConditions] = {
    ReactionInvalid: "not-acceptable",
    Forbidden: "forbidden",
    BadRequest: "bad-request",
    Unauthorized: "not-authorized",
    InvalidUserException: "item-not-found",
}


def tg_to_xmpp_errors(func: WrappedMethod) -> WrappedMethod:
    @functools.wraps(func)
    async def wrapped(*a, **ka):
        try:
            return await func(*a, **ka)
        except AuthKeyUnregistered:
            self: "Session" = a[0]
            await self.on_invalid_key()
        except (RPCError, InvalidUserException) as e:
            _raise(e, func)

    return wrapped


def tg_to_xmpp_errors_it(func: WrappedMethod) -> WrappedMethod:
    @functools.wraps(func)
    async def wrapped(*a, **ka):
        try:
            async for x in func(*a, **ka):
                yield x
        except AuthKeyUnregistered:
            self: "Session" = a[0]
            await self.on_invalid_key()
        except (RPCError, InvalidUserException) as e:
            _raise(e, func)

    return wrapped


def log_error_on_peer_id_invalid(func: WrappedMethod) -> WrappedMethod:
    """
    Decorator to log an error when a telegram event is ignored because of a
    PeerIdInvalid error. Unfortunately, because of slidge's design, if a telegram
    profile cannot be fetched, we need to raise an XMPPError to prevent filling the
    LegacyRoster with invalid user IDs.
    Ideally, we would need to let events propagate to XMPP even if the profile cannot
    be fetched, but this would require some serious refactoring in slidge core. It is
    not even clear whether this is actually achievable, how would we discriminate
    between "bogus user IDs" and "deleted accounts" since telegram does not explicitly
    make the difference?
    Part of the issue is related to MUCs, where we *need* a nickname and not just a user
    ID to translate a telegram event.
    """

    @functools.wraps(func)
    async def wrapped(self, *a, **ka):
        try:
            return await func(self, *a, **ka)
        except XMPPError as e:
            self.log.error(
                "%r in %s called with %s and %s", e.text, func.__name__, a, ka
            )

    return wrapped


def ignore_event_on_peer_id_invalid(func: WrappedMethod) -> WrappedMethod:
    """
    Decorator to silently drop telegram events related to PeerIdInvalid errors.
    This seems to be related to deleted telegram accounts. In some situations, we do not
    even want to log an error when this happens, eg, for message deletion events or
    cached reactions from deleted accounts, since consequences are negligible.
    """

    @functools.wraps(func)
    async def wrapped(self, *a, **ka):
        try:
            return await func(self, *a, **ka)
        except XMPPError as e:
            if e.condition == "item-not-found":
                return
            else:
                self.log.error(
                    "%r in %s called with %s and %s", e.text, func.__name__, a, ka
                )

    return wrapped


def _raise(e: RPCError | InvalidUserException, func: WrappedMethod):
    condition = _ERROR_MAP.get(type(e), "internal-server-error")
    raise XMPPError(
        condition, getattr(e, "MESSAGE", str(e.args)) + f" in '{func.__name__}'"
    )



from datetime import datetime, timedelta, timezone
from io import BytesIO
from typing import TYPE_CHECKING

from pyrogram.enums import ChatMemberStatus, ChatType
from pyrogram.errors import UserNotParticipant
from pyrogram.types import ChatMember, ChatPermissions, ChatPrivileges, Message
from pyrogram.utils import zero_datetime
from slidge import global_config
from slidge.group import LegacyBookmarks, LegacyMUC, LegacyParticipant, MucType
from slidge.util.types import Hat, HoleBound, MucAffiliation
from slixmpp.exceptions import XMPPError

from .avatar import SetAvatarMixin
from .contact import Contact
from .errors import tg_to_xmpp_errors, tg_to_xmpp_errors_it
from .reactions import ReactionsMixin
from .telegram import Client
from .text_entities import entities_to_xep_0393
from .tg_msg import TelegramMessageSenderMixin

if TYPE_CHECKING:
    from .session import Session


class Bookmarks(LegacyBookmarks[int, "MUC"]):
    session: "Session"

    @property
    def tg(self) -> Client:
        return self.session.tg

    @staticmethod
    async def legacy_id_to_jid_local_part(legacy_id: int):
        return "group" + str(legacy_id)

    async def jid_local_part_to_legacy_id(self, local_part: str):
        try:
            chat_id = int(local_part.replace("group", ""))
        except ValueError:
            raise XMPPError(
                "bad-request",
                (
                    "This does not look like a valid telegram ID, at least not for"
                    " slidge. Do not be like edhelas, do not attempt to join groups you"
                    " had joined through spectrum. "
                ),
            )
        return chat_id

    async def fill(self):
        dialog_it = self.tg.get_dialogs()
        assert dialog_it is not None
        async for dialog in dialog_it:
            if dialog.chat.type in (
                ChatType.GROUP,
                ChatType.SUPERGROUP,
                ChatType.CHANNEL,
            ):
                if (
                    dialog.top_message is not None
                    and dialog.top_message.left_chat_member is not None
                    and dialog.top_message.left_chat_member.is_self
                ):
                    # destroyed groups or groups that have been left
                    continue
                try:
                    muc = await self.by_legacy_id(dialog.chat.id)
                except XMPPError:
                    continue
                await muc.add_to_bookmarks(
                    auto_join=dialog.chat.type == ChatType.GROUP, pin=dialog.is_pinned
                )


class MUC(ReactionsMixin, SetAvatarMixin, LegacyMUC[int, int, "Participant", int]):
    session: "Session"
    legacy_id: int

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.pinned_message_ids = list[int]()

    @property
    def tg(self) -> Client:
        return self.session.tg

    @tg_to_xmpp_errors
    async def update_info(self):
        try:
            await self.tg.get_chat_member(self.legacy_id, "me")
        except UserNotParticipant:
            raise XMPPError(
                "subscription-required", "You are not a member of this group."
            )
        chat = await self.tg.get_chat(self.legacy_id)

        self.name = chat.title
        self.n_participants = chat.members_count

        if hasattr(chat, "description"):
            self.description = chat.description
        if hasattr(chat, "pinned_message"):
            self.set_tg_pinned_message(chat.pinned_message)

        if chat.type == ChatType.SUPERGROUP:
            self.type = MucType.CHANNEL_NON_ANONYMOUS
        elif chat.type == ChatType.CHANNEL:
            self.type = MucType.CHANNEL
            self.name += " (channel)"
        elif chat.type == ChatType.GROUP:
            self.type = MucType.GROUP
        else:
            raise XMPPError(
                "bad-request", f"This is not a telegram group, but a {chat.type}"
            )

        await self.update_chat_photo(chat.photo)

    @tg_to_xmpp_errors_it
    async def fill_participants(self):
        if self.type == MucType.CHANNEL:
            me = await self.get_user_participant()
            me.role = "visitor"
            yield me
            return

        me = None  # type:ignore
        me_member = None
        it = self.tg.get_chat_members(self.legacy_id, limit=100)
        assert it is not None
        async for member in it:
            if member.user is None:
                self.log.debug("Member but no user: %s", member)
                continue

            participant = await self.get_participant_by_legacy_id(member.user.id)
            participant.update_tg_member(member)
            if participant.is_user:
                me = participant
                me_member = member
                continue
            yield participant

        if me is None or me_member is None:
            me = await self.get_user_participant()
            me_member = await self.tg.get_chat_member(self.legacy_id, "me")
        me.update_tg_member(me_member)
        yield me

    @tg_to_xmpp_errors
    async def backfill(
        self,
        after: HoleBound | None = None,
        before: HoleBound | None = None,
    ):
        now = datetime.now()
        self.log.debug("Fetching history between %s and %s", after, before)
        it = self.tg.get_chat_history(
            chat_id=self.legacy_id,
            limit=0,
            offset_id=0 if before is None else before.id,  # type:ignore
            min_id=0 if after is None else after.id,  # type:ignore
        )
        assert it is not None
        async for msg in it:
            self.log.debug("Fetched history message from %s", msg.date)
            if (now - msg.date).days > global_config.MAM_MAX_DAYS:
                break
            sender, _ = await self.session.get_sender(msg)
            await sender.send_tg_msg(msg, archive_only=True)

    @tg_to_xmpp_errors
    async def on_set_affiliation(
        self,
        contact: Contact,  # type:ignore
        affiliation: MucAffiliation,
        reason: str | None,
        nickname: str | None,
    ):
        member = await self.tg.get_chat_member(self.legacy_id, contact.legacy_id)

        if affiliation == "outcast":
            await self._on_ban(member, contact)

        elif affiliation == "member":
            await self._on_set_member(member, contact)

        elif affiliation in ("admin", "owner"):
            await self._on_set_admin(member, contact)

    @tg_to_xmpp_errors
    async def on_kick(
        self,
        contact: Contact,  # type:ignore
        reason: str | None,
        nickname: str | None,
    ):
        member = await self.tg.get_chat_member(self.legacy_id, contact.legacy_id)
        await self._on_ban(
            member, contact, datetime.now(tz=timezone.utc) + timedelta(minutes=5)
        )

    async def _on_ban(
        self, member: ChatMember, contact: Contact, until: datetime = zero_datetime()
    ) -> None:
        if member.status == ChatMemberStatus.BANNED:
            raise XMPPError(
                "bad-request",
                f"{contact.name} is already banned from {self.name}.",
            )
        elif member.status == ChatMemberStatus.OWNER:
            raise XMPPError(
                "bad-request",
                f"{contact.name} is the owner of {self.name}.",
            )
        elif member.status == ChatMemberStatus.LEFT:
            raise XMPPError(
                "bad-request",
                f"{contact.name} has already left {self.name}.",
            )
        if await self.tg.ban_chat_member(self.legacy_id, contact.legacy_id, until):
            participant = await self.get_participant_by_legacy_id(contact.legacy_id)
            participant.affiliation = "outcast"
            participant.role = "none"
        else:
            raise XMPPError("internal-server-error")

    async def _on_set_member(self, member: ChatMember, contact: Contact) -> None:
        if member.status == ChatMemberStatus.BANNED:
            success = await self.tg.unban_chat_member(self.legacy_id, contact.legacy_id)
        elif member.status == ChatMemberStatus.RESTRICTED:
            await self.tg.restrict_chat_member(
                self.legacy_id,
                contact.legacy_id,
                permissions=ChatPermissions(all_perms=True),
            )
            success = True
        elif member.status == ChatMemberStatus.LEFT:
            success = await self.tg.add_chat_members(
                self.legacy_id, [contact.legacy_id]
            )
        elif member.status == ChatMemberStatus.ADMINISTRATOR:
            if self.type == MucType.GROUP:
                success = await self.tg.edit_chat_admin(
                    self.legacy_id, contact.legacy_id, False
                )
            else:
                success = await self.tg.promote_chat_member(
                    self.legacy_id,
                    contact.legacy_id,
                    _NO_PRIVILEGES,
                )
        elif member.status == ChatMemberStatus.MEMBER:
            raise XMPPError(
                "bad-request",
                f"{contact.name} is already a member of {self.name}",
            )
        elif member.status == ChatMemberStatus.OWNER:
            raise XMPPError("bad-request", "The chat owner cannot be demoted")

        if success:
            participant = await self.get_participant_by_legacy_id(contact.legacy_id)
            participant.affiliation = "member"
            participant.role = "participant"
        else:
            raise XMPPError("internal-server-error")

    async def _on_set_admin(self, member: ChatMember, contact: Contact) -> None:
        if member.status in (ChatMemberStatus.OWNER, ChatMemberStatus.ADMINISTRATOR):
            raise XMPPError(
                "bad-request", f"This chat member is already {member.status}"
            )
        elif member.status == ChatMemberStatus.LEFT:
            if not await self.tg.add_chat_members(self.legacy_id, [contact.legacy_id]):
                raise XMPPError("internal-server-error")

        if self.type == MucType.GROUP:
            success = await self.tg.edit_chat_admin(
                self.legacy_id, contact.legacy_id, True
            )
        else:
            success = await self.tg.promote_chat_member(
                self.legacy_id, contact.legacy_id, _ALL_PRIVILEGES
            )

        if success:
            participant = await self.get_participant_by_legacy_id(contact.legacy_id)
            participant.affiliation = "admin"
            participant.role = "moderator"
        else:
            raise XMPPError("internal-server-error")

    @tg_to_xmpp_errors
    async def on_set_config(
        self,
        name: str | None,
        description: str | None,
    ):
        name = name or ""
        description = description or ""
        if name != self.name:
            if await self.tg.set_chat_title(self.legacy_id, name):
                self.name = name
            else:
                raise XMPPError(
                    "internal-server-error", "While trying to rename %s", self.name
                )
        if description != self.description:
            if await self.tg.set_chat_description(self.legacy_id, description):
                self.description = description
            else:
                raise XMPPError(
                    "internal-server-error",
                    "While trying to change the description of %s",
                    self.name,
                )

    @tg_to_xmpp_errors
    async def on_destroy_request(self, reason: str | None):
        if self.type == MucType.CHANNEL_NON_ANONYMOUS:
            success = await self.tg.delete_supergroup(self.legacy_id)
        elif self.type == MucType.CHANNEL:
            success = await self.tg.delete_channel(self.legacy_id)
        else:
            await self.tg.leave_chat(self.legacy_id, delete=True)
            success = True

        if not success:
            raise XMPPError("internal-server-error", "The group could not be deleted")

    @tg_to_xmpp_errors
    async def on_avatar(self, data: bytes | None, mime: str | None) -> None:
        if not data:
            await self.tg.delete_chat_photo(self.legacy_id)
            return

        await self.tg.set_chat_photo(self.legacy_id, photo=BytesIO(data))

    async def on_set_subject(self, subject: str) -> None:
        raise XMPPError(
            "feature-not-implemented",
            "Telegram groups don't have subject but pinned messages, "
            "and they cannot be set from XMPP.",
        )

    def set_tg_pinned_message(self, message: Message | None = None) -> None:
        if message is None:
            self.subject = ""
            self.subject_date = None
            self.subject_setter = None
            return

        self.pinned_message_ids.append(message.id)
        if message.date is None:
            self.subject_date = None
        else:
            if message.edit_date is None:
                self.subject_date = message.date.replace(tzinfo=timezone.utc)
            else:
                self.subject_date = message.edit_date.replace(tzinfo=timezone.utc)
        if message.from_user is None:
            self.subject_setter = None
        else:
            self.subject_setter = (
                message.from_user.username or message.from_user.full_name
            )

        if message.text is not None:
            assert self.tg.me is not None
            self.subject = entities_to_xep_0393(
                message.text, message.entities, self.tg.me.id, self.user_nick
            )
            return

        if message.caption is not None:
            assert self.tg.me is not None
            self.subject = entities_to_xep_0393(
                message.caption, message.caption_entities, self.tg.me.id, self.user_nick
            )
            return

        self.subject = (
            f"The last pinned message is an attachment without caption: {message.media}."
            "This is unsupported by slidgram."
        )

    async def set_tg_pinned_message_ids(
        self, message_ids: list[int], pinned: bool | None
    ) -> None:
        if pinned:
            self.pinned_message_ids.extend(message_ids)
        else:
            for message_id in message_ids:
                try:
                    self.pinned_message_ids.remove(message_id)
                except ValueError:
                    self.log.warning("Unpinned a message we didn't know?")
        if not self.pinned_message_ids:
            self.set_tg_pinned_message()
            return

        message_id = max(self.pinned_message_ids)
        message = await self.tg.get_messages(self.legacy_id, message_id)
        assert isinstance(message, Message)
        self.set_tg_pinned_message(message)

    def serialize_extra_attributes(self) -> dict:
        return {"pinned_messages": self.pinned_message_ids}

    def deserialize_extra_attributes(self, data: dict) -> None:
        self.pinned_message_ids = data.get("pinned_messages", [])


class Participant(TelegramMessageSenderMixin, LegacyParticipant):
    muc: MUC

    async def send_tg_msg(
        self, message: Message, carbon=False, correction=False, archive_only=False
    ) -> None:
        if message.new_chat_photo is not None:
            await self.muc.update_chat_photo(message.new_chat_photo)
        if message.delete_chat_photo is not None and message.delete_chat_photo:
            await self.muc.set_avatar(None)

        if message.new_chat_title is not None:
            self.muc.name = message.new_chat_title
        if message.new_chat_members is not None:
            for tg_user in message.new_chat_members:
                await self.muc.get_participant_by_legacy_id(tg_user.id)
        if (message.group_chat_created is not None and message.group_chat_created) or (
            message.supergroup_chat_created and message.supergroup_chat_created
        ):
            await self.muc.add_to_bookmarks()

        await super().send_tg_msg(
            message, archive_only=archive_only, correction=correction
        )

    def update_tg_member(self, member: ChatMember):
        if member.status == ChatMemberStatus.OWNER:
            self.affiliation = "owner"
            self.role = "moderator"
        elif member.status == ChatMemberStatus.ADMINISTRATOR:
            if is_owner_privileges(member.privileges):
                self.affiliation = "owner"
            else:
                self.affiliation = "admin"
            self.role = "moderator"
        elif member.status == ChatMemberStatus.RESTRICTED:
            self.set_hats([Hat("https://slidge.im/slidgram/hats", "restricted")])
            self.role = "visitor"
        elif member.status == ChatMemberStatus.LEFT:
            self.leave()
        elif member.status == ChatMemberStatus.BANNED:
            self.affiliation = "outcast"
            self.role = "none"
            self.leave()

        if member.custom_title is not None:
            self.set_hats([Hat("https://slidge.im/slidgram/hats", member.custom_title)])


def is_owner_privileges(privileges: ChatPrivileges | None) -> bool:
    return privileges is not None and privileges.can_change_info


_NO_PRIVILEGES = ChatPrivileges(
    can_manage_chat=False,
    can_delete_messages=False,
    can_manage_video_chats=False,  # Groups and supergroups only
    can_restrict_members=False,
    can_promote_members=False,
    can_change_info=False,
    can_post_messages=False,  # Channels only
    can_edit_messages=False,  # Channels only
    can_invite_users=False,
    can_pin_messages=False,  # Groups and supergroups only
    can_manage_topics=False,  # supergroups only.
    can_post_stories=False,  # Channels only
    can_edit_stories=False,  # Channels only
    can_delete_stories=False,  # Channels only
)

_ALL_PRIVILEGES = ChatPrivileges(
    can_manage_chat=True,
    can_delete_messages=True,
    can_manage_video_chats=True,  # Groups and supergroups only
    can_restrict_members=True,
    can_promote_members=True,
    can_change_info=True,
    can_post_messages=True,  # Channels only
    can_edit_messages=True,  # Channels only
    can_invite_users=True,
    can_pin_messages=True,  # Groups and supergroups only
    can_manage_topics=True,  # supergroups only.
    can_post_stories=True,  # Channels only
    can_edit_stories=True,  # Channels only
    can_delete_stories=True,  # Channels only
)
